import msgspec
from typing import Dict, List
from decimal import Decimal

from nexustrader.error import PositionModeError

from nexustrader.core.nautilius_core import LiveClock, MessageBus
from nexustrader.base import OrderManagementSystem
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.core.cache import AsyncCache
from nexustrader.schema import (
    Order,
    Position,
    BatchOrderSubmit,
)
from nexustrader.constants import (
    OrderSide,
    OrderStatus,
    ExchangeType,
    OrderType,
    TimeInForce,
    PositionSide,
    TriggerType,
)
from nexustrader.exchange.bybit.schema import (
    BybitWsMessageGeneral,
    BybitWsApiGeneralMsg,
    BybitWsOrderMsg,
    BybitMarket,
    BybitWsPositionMsg,
    BybitWsAccountWalletMsg,
    BybitWalletBalanceResponse,
    BybitPositionStruct,
    BybitWsApiOrderMsg,
)
from nexustrader.exchange.bybit.rest_api import BybitApiClient
from nexustrader.exchange.bybit.websockets import BybitWSClient, BybitWSApiClient
from nexustrader.exchange.bybit.constants import (
    BybitAccountType,
    BybitEnumParser,
    BybitProductType,
    BybitOrderType,
    BybitTimeInForce,
)


class BybitOrderManagementSystem(OrderManagementSystem):
    _ws_client: BybitWSClient
    _account_type: BybitAccountType
    _market: Dict[str, BybitMarket]
    _market_id: Dict[str, str]
    _api_client: BybitApiClient

    def __init__(
        self,
        account_type: BybitAccountType,
        api_key: str,
        secret: str,
        market: Dict[str, BybitMarket],
        market_id: Dict[str, str],
        registry: OrderRegistry,
        cache: AsyncCache,
        api_client: BybitApiClient,
        exchange_id: ExchangeType,
        clock: LiveClock,
        msgbus: MessageBus,
        task_manager: TaskManager,
        enable_rate_limit: bool,
    ):
        super().__init__(
            account_type=account_type,
            market=market,
            market_id=market_id,
            registry=registry,
            cache=cache,
            api_client=api_client,
            ws_client=BybitWSClient(
                account_type=account_type,
                handler=self._ws_msg_handler,
                task_manager=task_manager,
                clock=clock,
                api_key=api_key,
                secret=secret,
            ),
            exchange_id=exchange_id,
            clock=clock,
            msgbus=msgbus,
        )

        self._ws_api_client = BybitWSApiClient(
            account_type=account_type,
            api_key=api_key,
            secret=secret,
            handler=self._ws_api_msg_handler,
            task_manager=task_manager,
            clock=clock,
            enable_rate_limit=enable_rate_limit,
        )

        self._ws_api_msg_general_decoder = msgspec.json.Decoder(BybitWsApiGeneralMsg)
        self._ws_api_msg_order_decoder = msgspec.json.Decoder(BybitWsApiOrderMsg)
        self._ws_msg_general_decoder = msgspec.json.Decoder(BybitWsMessageGeneral)
        self._ws_msg_order_update_decoder = msgspec.json.Decoder(BybitWsOrderMsg)
        self._ws_msg_position_decoder = msgspec.json.Decoder(BybitWsPositionMsg)
        self._ws_msg_wallet_decoder = msgspec.json.Decoder(BybitWsAccountWalletMsg)

    def _parse_order_create(self, raw: bytes):
        msg = self._ws_api_msg_order_decoder.decode(raw)
        oid = msg.oid

        tmp_order = self._registry.get_tmp_order(oid)
        if not tmp_order:
            return
        symbol = tmp_order.symbol
        amount = tmp_order.amount
        type = tmp_order.type
        price = tmp_order.price
        side = tmp_order.side
        time_in_force = tmp_order.time_in_force
        reduce_only = tmp_order.reduce_only
        ts = self._clock.timestamp_ms()
        if msg.is_success:
            ordId = msg.data.orderId
            self._log.debug(f"[{symbol}] new order success: oid: {oid} eid: {ordId}")
            order = Order(
                exchange=self._exchange_id,
                symbol=symbol,
                oid=oid,
                eid=ordId,
                side=side,
                status=OrderStatus.PENDING,
                amount=amount,
                timestamp=ts,
                type=type,
                price=price,
                time_in_force=time_in_force,
                reduce_only=reduce_only,
            )
            self.order_status_update(order)  # INITIALIZED -> PENDING
        else:
            self._log.error(f"[{symbol}] new order failed: oid: {oid} {msg.error_msg}")
            order = Order(
                exchange=self._exchange_id,
                symbol=symbol,
                oid=oid,
                status=OrderStatus.FAILED,
                amount=amount,
                timestamp=ts,
                side=side,
                type=type,
                price=price,
                time_in_force=time_in_force,
                reduce_only=reduce_only,
            )
            self.order_status_update(order)

    def _parse_order_cancel(self, raw: bytes):
        msg = self._ws_api_msg_order_decoder.decode(raw)
        oid = msg.oid
        tmp_order = self._registry.get_tmp_order(oid)
        if not tmp_order:
            return
        symbol = tmp_order.symbol
        amount = tmp_order.amount
        side = tmp_order.side
        type = tmp_order.type
        price = tmp_order.price
        time_in_force = tmp_order.time_in_force
        reduce_only = tmp_order.reduce_only
        ts = self._clock.timestamp_ms()
        if msg.is_success:
            ordId = msg.data.orderId
            self._log.debug(
                f"[{symbol}] canceling order success: oid: {oid} id: {ordId}"
            )
            order = Order(
                exchange=self._exchange_id,
                symbol=symbol,
                oid=oid,
                eid=ordId,
                side=side,
                status=OrderStatus.CANCELING,
                amount=amount,
                type=type,
                price=price,
                time_in_force=time_in_force,
                timestamp=ts,
                reduce_only=reduce_only,
            )
            self.order_status_update(order)
        else:
            self._log.error(
                f"[{symbol}] canceling order failed: oid: {oid} {msg.error_msg}"
            )
            order = Order(
                exchange=self._exchange_id,
                symbol=symbol,
                oid=oid,
                status=OrderStatus.CANCEL_FAILED,
                amount=amount,
                side=side,
                type=type,
                price=price,
                time_in_force=time_in_force,
                timestamp=ts,
                reduce_only=reduce_only,
            )
            self.order_status_update(order)

    def _ws_api_msg_handler(self, raw: bytes):
        try:
            ws_msg = self._ws_api_msg_general_decoder.decode(raw)
            if ws_msg.is_pong:
                self._ws_api_client._transport.notify_user_specific_pong_received()
                self._log.debug(f"Pong received {str(ws_msg)}")
                return
            elif ws_msg.is_order_create:
                self._parse_order_create(raw)
            elif ws_msg.is_order_cancel:
                self._parse_order_cancel(raw)
        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding message: {str(raw)} {e}")

    def _ws_msg_handler(self, raw: bytes):
        try:
            ws_msg = self._ws_msg_general_decoder.decode(raw)
            if ws_msg.op == "pong":
                self._ws_client._transport.notify_user_specific_pong_received()
                self._log.debug(f"Pong received {str(ws_msg)}")
                return
            if ws_msg.success is False:
                self._log.error(f"WebSocket error: {ws_msg}")
                return
            if "order" in ws_msg.topic:
                self._parse_order_update(raw)
            elif "position" in ws_msg.topic:
                self._parse_position_update(raw)
            elif "wallet" == ws_msg.topic:
                self._parse_wallet_update(raw)
        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding message: {str(raw)} {e}")

    def _get_category(self, market: BybitMarket):
        if market.spot:
            return "spot"
        elif market.linear:
            return "linear"
        elif market.inverse:
            return "inverse"
        else:
            raise ValueError(f"Unsupported market type: {market.type}")

    async def cancel_order_ws(self, oid: str, symbol: str, **kwargs):
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        id = market.id
        category = self._get_category(market)
        params = {
            "category": category,
            "symbol": id,
            "orderLinkId": oid,
            **kwargs,
        }

        await self._ws_api_client.cancel_order(id=oid, **params)

    async def cancel_order(self, oid: str, symbol: str, **kwargs):
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        id = market.id
        category = self._get_category(market)
        params = {
            "category": category,
            "symbol": id,
            "orderLinkId": oid,
            **kwargs,
        }
        try:
            res = await self._api_client.post_v5_order_cancel(**params)
            order = Order(
                oid=oid,
                eid=res.result.orderId,
                exchange=self._exchange_id,
                timestamp=res.time,
                symbol=symbol,
                status=OrderStatus.CANCELING,
            )
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error canceling order: {error_msg} params: {str(params)}")
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=symbol,
                status=OrderStatus.CANCEL_FAILED,
            )
        self.order_status_update(order)

    def _init_account_balance(self):
        res: BybitWalletBalanceResponse = (
            self._api_client.get_v5_account_wallet_balance(account_type="UNIFIED")
        )
        for result in res.result.list:
            self._cache._apply_balance(self._account_type, result.parse_to_balances())

    def _get_all_positions_list(
        self, category: BybitProductType, settle_coin: str | None = None
    ) -> list[BybitPositionStruct]:
        all_positions = []
        next_page_cursor = ""

        while True:
            res = self._api_client.get_v5_position_list(
                category=category.value,
                settleCoin=settle_coin,
                limit=200,
                cursor=next_page_cursor,
            )

            all_positions.extend(res.result.list)

            # If there's no next page cursor, we've reached the end
            if not res.result.nextPageCursor:
                break

            next_page_cursor = res.result.nextPageCursor

        return all_positions

    def _init_position(self):
        res_linear_usdt = self._get_all_positions_list(
            BybitProductType.LINEAR, settle_coin="USDT"
        )
        res_linear_usdc = self._get_all_positions_list(
            BybitProductType.LINEAR, settle_coin="USDC"
        )
        res_inverse = self._get_all_positions_list(BybitProductType.INVERSE)

        self._apply_cache_position(res_linear_usdt, BybitProductType.LINEAR)
        self._apply_cache_position(res_linear_usdc, BybitProductType.LINEAR)
        self._apply_cache_position(res_inverse, BybitProductType.INVERSE)

        self._cache.get_all_positions()

    def _position_mode_check(self):
        # NOTE: no need to implement this for bybit, we do position mode check in _get_all_positions_list
        pass

    def _apply_cache_position(
        self, positions: list[BybitPositionStruct], category: BybitProductType
    ):
        for result in positions:
            side = result.side.parse_to_position_side()
            if side == PositionSide.FLAT:
                signed_amount = Decimal(0)
                side = None
            elif side == PositionSide.LONG:
                signed_amount = Decimal(result.size)
            elif side == PositionSide.SHORT:
                signed_amount = -Decimal(result.size)

            if category.is_inverse:
                id = result.symbol + "_inverse"
            elif category.is_linear:
                id = result.symbol + "_linear"

            symbol = self._market_id[id]

            if not result.positionIdx.is_one_way_mode():
                raise PositionModeError(
                    f"Please Set Position Mode to `One-Way Mode` in Bybit App for {symbol}"
                )

            position = Position(
                symbol=symbol,
                exchange=self._exchange_id,
                side=side,
                signed_amount=signed_amount,
                entry_price=float(result.avgPrice),
                unrealized_pnl=float(result.unrealisedPnl),
                realized_pnl=float(result.cumRealisedPnl),
            )
            self._cache._apply_position(position)

    async def create_tp_sl_order(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce | None = TimeInForce.GTC,
        tp_order_type: OrderType | None = None,
        tp_trigger_price: Decimal | None = None,
        tp_price: Decimal | None = None,
        tp_trigger_type: TriggerType | None = TriggerType.LAST_PRICE,
        sl_order_type: OrderType | None = None,
        sl_trigger_price: Decimal | None = None,
        sl_price: Decimal | None = None,
        sl_trigger_type: TriggerType | None = TriggerType.LAST_PRICE,
        **kwargs,
    ) -> Order:
        """Create a take profit and stop loss order"""
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        id = market.id

        category = self._get_category(market)

        params = {
            "category": category,
            "symbol": id,
            "side": BybitEnumParser.to_bybit_order_side(side).value,
            "qty": str(amount),
            "orderLinkId": oid,
        }

        if type.is_limit:
            if not price:
                raise ValueError("Price is required for limit order")
            params["price"] = str(price)
            params["order_type"] = BybitOrderType.LIMIT.value
            params["timeInForce"] = BybitEnumParser.to_bybit_time_in_force(
                time_in_force
            ).value
        elif type.is_post_only:
            if not price:
                raise ValueError("Price is required for post-only order")
            params["order_type"] = BybitOrderType.LIMIT.value
            params["price"] = str(price)
            params["timeInForce"] = BybitTimeInForce.POST_ONLY.value
        elif type == OrderType.MARKET:
            params["order_type"] = BybitOrderType.MARKET.value

        if market.spot:
            params["marketUnit"] = "baseCoin"

        if tp_order_type:
            params["takeProfit"] = str(tp_trigger_price)
            params["triggerBy"] = BybitEnumParser.to_bybit_trigger_type(
                tp_trigger_type
            ).value
            if tp_order_type.is_limit:
                if not tp_price:
                    raise ValueError("Price is required for limit take profit order")
                params["tpOrderType"] = BybitOrderType.LIMIT.value
                params["tpLimitPrice"] = str(tp_price)

        if sl_order_type:
            params["stopLoss"] = str(sl_trigger_price)
            params["triggerBy"] = BybitEnumParser.to_bybit_trigger_type(
                sl_trigger_type
            ).value
            if sl_order_type.is_limit:
                if not sl_price:
                    raise ValueError("Price is required for limit stop loss order")
                params["slOrderType"] = BybitOrderType.LIMIT.value
                params["slLimitPrice"] = str(sl_price)

        if not market.spot:
            tpslMode = kwargs.pop("tpslMode", "Partial")
            params["tpslMode"] = tpslMode

        params.update(kwargs)

        try:
            res = await self._api_client.post_v5_order_create(**params)

            order = Order(
                exchange=self._exchange_id,
                eid=res.result.orderId,
                oid=oid,
                timestamp=int(res.time),
                symbol=symbol,
                type=type,
                side=side,
                amount=amount,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                status=OrderStatus.PENDING,
                filled=Decimal(0),
                remaining=amount,
                reduce_only=False,
            )
            return order
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error creating order: {error_msg} params: {str(params)}")
            order = Order(
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=symbol,
                oid=oid,
                type=type,
                side=side,
                amount=amount,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                status=OrderStatus.FAILED,
                filled=Decimal(0),
                remaining=amount,
            )
            return order

    async def create_batch_orders(self, orders: List[BatchOrderSubmit]):
        if not orders:
            raise ValueError("No orders provided for batch submission")

        # Get category from first order
        first_market = self._market.get(orders[0].symbol)
        if not first_market:
            raise ValueError(f"Symbol {orders[0].symbol} not found in market")
        category = self._get_category(first_market)

        batch_orders = []
        for order in orders:
            market = self._market.get(order.symbol)
            if not market:
                raise ValueError(f"Symbol {order.symbol} not found in market")
            id = market.id

            params = {
                "symbol": id,
                "side": BybitEnumParser.to_bybit_order_side(order.side).value,
                "qty": str(order.amount),
                "orderLinkId": order.oid,
            }
            if order.type.is_limit:
                if not order.price:
                    raise ValueError("Price is required for limit order")
                params["orderType"] = BybitOrderType.LIMIT.value
                params["price"] = str(order.price)
                params["timeInForce"] = BybitEnumParser.to_bybit_time_in_force(
                    order.time_in_force
                ).value
            elif order.type.is_post_only:
                if not order.price:
                    raise ValueError("Price is required for limit order")
                params["orderType"] = BybitOrderType.LIMIT.value
                params["price"] = str(order.price)
                params["timeInForce"] = BybitTimeInForce.POST_ONLY.value
            elif order.type == OrderType.MARKET:
                params["orderType"] = BybitOrderType.MARKET.value

            if order.reduce_only:
                params["reduceOnly"] = True
            if market.spot:
                params["marketUnit"] = "baseCoin"
            params.update(order.kwargs)
            batch_orders.append(params)

        try:
            res = await self._api_client.post_v5_order_create_batch(
                category=category, request=batch_orders
            )
            for order, res_order, res_ext in zip(
                orders, res.result.list, res.retExtInfo.list
            ):
                if res_ext.code == 0:
                    res_batch_order = Order(
                        exchange=self._exchange_id,
                        oid=order.oid,
                        eid=res_order.orderId,
                        timestamp=int(res_order.createAt),
                        symbol=order.symbol,
                        type=order.type,
                        side=order.side,
                        amount=order.amount,
                        price=float(order.price) if order.price else None,
                        time_in_force=order.time_in_force,
                        status=OrderStatus.PENDING,
                        filled=Decimal(0),
                        remaining=order.amount,
                        reduce_only=order.reduce_only,
                    )
                else:
                    res_batch_order = Order(
                        exchange=self._exchange_id,
                        timestamp=self._clock.timestamp_ms(),
                        symbol=order.symbol,
                        type=order.type,
                        oid=order.oid,
                        side=order.side,
                        amount=order.amount,
                        price=float(order.price) if order.price else None,
                        time_in_force=order.time_in_force,
                        status=OrderStatus.FAILED,
                        filled=Decimal(0),
                        remaining=order.amount,
                        reduce_only=order.reduce_only,
                    )
                    self._log.error(
                        f"Failed to place order for {order.symbol}: {res_ext.msg} code: {res_ext.code} {order.oid}"
                    )
                self.order_status_update(res_batch_order)
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error creating batch orders: {error_msg}")
            for order in orders:
                res_batch_order = Order(
                    exchange=self._exchange_id,
                    timestamp=self._clock.timestamp_ms(),
                    symbol=order.symbol,
                    oid=order.oid,
                    type=order.type,
                    side=order.side,
                    amount=order.amount,
                    price=float(order.price) if order.price else None,
                    time_in_force=order.time_in_force,
                    status=OrderStatus.FAILED,
                    filled=Decimal(0),
                    remaining=order.amount,
                    reduce_only=order.reduce_only,
                )
                self.order_status_update(res_batch_order)

    async def create_order(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce = TimeInForce.GTC,
        reduce_only: bool = False,
        **kwargs,
    ):
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        id = market.id

        category = self._get_category(market)

        params = {
            "category": category,
            "symbol": id,
            "side": BybitEnumParser.to_bybit_order_side(side).value,
            "qty": str(amount),
            "orderLinkId": oid,
        }

        if type.is_limit:
            if not price:
                raise ValueError("Price is required for limit order")
            params["price"] = str(price)
            params["order_type"] = BybitOrderType.LIMIT.value
            params["timeInForce"] = BybitEnumParser.to_bybit_time_in_force(
                time_in_force
            ).value
        elif type.is_post_only:
            if not price:
                raise ValueError("Price is required for post-only order")
            params["order_type"] = BybitOrderType.LIMIT.value
            params["price"] = str(price)
            params["timeInForce"] = BybitTimeInForce.POST_ONLY.value
        elif type == OrderType.MARKET:
            params["order_type"] = BybitOrderType.MARKET.value

        # if position_side:
        #     params["positionIdx"] = BybitEnumParser.to_bybit_position_side(
        #         position_side
        #     ).value
        if reduce_only:
            params["reduceOnly"] = True
        if market.spot:
            params["marketUnit"] = "baseCoin"

        params.update(kwargs)

        try:
            res = await self._api_client.post_v5_order_create(**params)
            order = Order(
                exchange=self._exchange_id,
                eid=res.result.orderId,
                oid=oid,
                timestamp=int(res.time),
                symbol=symbol,
                type=type,
                side=side,
                amount=amount,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                # position_side=position_side,
                status=OrderStatus.PENDING,
                filled=Decimal(0),
                remaining=amount,
                reduce_only=reduce_only,
            )
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error creating order: {error_msg} params: {str(params)}")
            order = Order(
                exchange=self._exchange_id,
                oid=oid,
                timestamp=self._clock.timestamp_ms(),
                symbol=symbol,
                type=type,
                side=side,
                amount=amount,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                # position_side=position_side,
                status=OrderStatus.FAILED,
                filled=Decimal(0),
                remaining=amount,
                reduce_only=reduce_only,
            )
        self.order_status_update(order)

    async def create_order_ws(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce = TimeInForce.GTC,
        reduce_only: bool = False,
        **kwargs,
    ):
        self._registry.register_tmp_order(
            order=Order(
                oid=oid,
                exchange=self._exchange_id,
                symbol=symbol,
                status=OrderStatus.INITIALIZED,
                amount=amount,
                type=type,
                side=side,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                reduce_only=reduce_only,
                timestamp=self._clock.timestamp_ms(),
            )
        )
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        id = market.id

        category = self._get_category(market)

        params = {
            "category": category,
            "symbol": id,
            "side": BybitEnumParser.to_bybit_order_side(side).value,
            "qty": str(amount),
            "orderLinkId": oid,
        }

        if type.is_limit:
            if not price:
                raise ValueError("Price is required for limit order")
            params["price"] = str(price)
            params["orderType"] = BybitOrderType.LIMIT.value
            params["timeInForce"] = BybitEnumParser.to_bybit_time_in_force(
                time_in_force
            ).value
        elif type.is_post_only:
            if not price:
                raise ValueError("Price is required for post-only order")
            params["orderType"] = BybitOrderType.LIMIT.value
            params["price"] = str(price)
            params["timeInForce"] = BybitTimeInForce.POST_ONLY.value
        elif type == OrderType.MARKET:
            params["orderType"] = BybitOrderType.MARKET.value

        # if position_side:
        #     params["positionIdx"] = BybitEnumParser.to_bybit_position_side(
        #         position_side
        #     ).value
        if reduce_only:
            params["reduceOnly"] = True
        if market.spot:
            params["marketUnit"] = "baseCoin"

        params.update(kwargs)

        await self._ws_api_client.create_order(id=oid, **params)

    async def cancel_all_orders(self, symbol: str) -> bool:
        try:
            market = self._market.get(symbol)
            if not market:
                raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
            symbol = market.id

            category = self._get_category(market)

            params = {
                "category": category,
                "symbol": symbol,
            }

            await self._api_client.post_v5_order_cancel_all(**params)
            return True
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(
                f"Error canceling all orders: {error_msg} params: {str(params)}"
            )
            return False

    async def modify_order(
        self,
        oid: str,
        symbol: str,
        side: OrderSide | None = None,
        price: Decimal | None = None,
        amount: Decimal | None = None,
        **kwargs,
    ):
        # NOTE: side is not supported for modify order
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        id = market.id

        category = self._get_category(market)
        params = {
            "category": category,
            "symbol": id,
            "orderLinkId": oid,
            "price": str(price) if price else None,
            "qty": str(amount) if amount else None,
            **kwargs,
        }

        try:
            res = await self._api_client.post_v5_order_amend(**params)
            order = Order(
                exchange=self._exchange_id,
                eid=res.result.orderId,
                oid=oid,
                timestamp=int(res.time),
                symbol=symbol,
                status=OrderStatus.PENDING,
                filled=Decimal(0),
                price=float(price) if price else None,
                remaining=amount,
            )
            return order
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error modifying order: {error_msg} params: {str(params)}")
            order = Order(
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=symbol,
                oid=oid,
                status=OrderStatus.FAILED,
                filled=Decimal(0),
                remaining=amount,
                price=float(price) if price else None,
            )
            return order

    def _parse_order_update(self, raw: bytes):
        order_msg = self._ws_msg_order_update_decoder.decode(raw)
        self._log.debug(f"Order update: {str(order_msg)}")
        for data in order_msg.data:
            category = data.category
            if category.is_spot:
                id = data.symbol + "_spot"
            elif category.is_linear:
                id = data.symbol + "_linear"
            elif category.is_inverse:
                id = data.symbol + "_inverse"
            symbol = self._market_id[id]

            order = Order(
                exchange=self._exchange_id,
                symbol=symbol,
                status=BybitEnumParser.parse_order_status(data.orderStatus),
                eid=data.orderId,
                oid=data.orderLinkId,
                timestamp=int(data.updatedTime),
                type=BybitEnumParser.parse_order_type(data.orderType, data.timeInForce),
                side=BybitEnumParser.parse_order_side(data.side),
                time_in_force=BybitEnumParser.parse_time_in_force(data.timeInForce),
                price=float(data.price),
                average=float(data.avgPrice) if data.avgPrice else None,
                amount=Decimal(data.qty),
                filled=Decimal(data.cumExecQty),
                remaining=Decimal(data.qty) - Decimal(data.cumExecQty),
                fee=Decimal(data.cumExecFee),
                fee_currency=data.feeCurrency,
                cum_cost=Decimal(data.cumExecValue),
                reduce_only=data.reduceOnly,
                position_side=BybitEnumParser.parse_position_side(data.positionIdx),
            )

            self.order_status_update(order)

    def _parse_position_update(self, raw: bytes):
        position_msg = self._ws_msg_position_decoder.decode(raw)
        self._log.debug(f"Position update: {str(position_msg)}")

        for data in position_msg.data:
            category = data.category
            if category.is_linear:  # only linear/inverse/ position is supported
                id = data.symbol + "_linear"
            elif category.is_inverse:
                id = data.symbol + "_inverse"
            symbol = self._market_id[id]

            side = data.side.parse_to_position_side()
            if side == PositionSide.LONG:
                signed_amount = Decimal(data.size)
            elif side == PositionSide.SHORT:
                signed_amount = -Decimal(data.size)
            else:
                side = None
                signed_amount = Decimal(0)

            position = Position(
                symbol=symbol,
                exchange=self._exchange_id,
                side=side,
                signed_amount=signed_amount,
                entry_price=float(data.entryPrice),
                unrealized_pnl=float(data.unrealisedPnl),
                realized_pnl=float(data.cumRealisedPnl),
            )

            self._cache._apply_position(position)

    def _parse_wallet_update(self, raw: bytes):
        wallet_msg = self._ws_msg_wallet_decoder.decode(raw)
        self._log.debug(f"Wallet update: {str(wallet_msg)}")

        for data in wallet_msg.data:
            balances = data.parse_to_balances()
            self._cache._apply_balance(self._account_type, balances)



import asyncio
import msgspec
from abc import ABC, abstractmethod
from typing import Any
from typing import Callable, Literal


from nexustrader.core.entity import TaskManager
from picows import (
    ws_connect,
    WSFrame,
    WSTransport,
    WSListener,
    WSMsgType,
    WSAutoPingStrategy,
    # PICOWS_DEBUG_LL,
)
from nexustrader.core.nautilius_core import LiveClock, Logger

# import logging

# file_handler = logging.FileHandler('.log/picows.log')
# file_handler.setLevel(PICOWS_DEBUG_LL)

# formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# file_handler.setFormatter(formatter)

# picows_logger = logging.getLogger("picows")
# picows_logger.setLevel(PICOWS_DEBUG_LL)
# picows_logger.addHandler(file_handler)


class Listener(WSListener):
    """WebSocket listener implementation that handles connection events and message frames.

    Inherits from picows.WSListener to provide WebSocket event handling functionality.
    """

    def __init__(self, callback, logger, specific_ping_msg=None, *args, **kwargs):
        """Initialize the WebSocket listener.

        Args:
            logger: Logger instance for logging events
            specific_ping_msg: Optional custom ping message
        """
        super().__init__(*args, **kwargs)
        self._log = logger
        self._specific_ping_msg: bytes = specific_ping_msg
        self._callback = callback

    def send_user_specific_ping(self, transport: WSTransport) -> None:
        """Send a custom ping message or default ping frame.

        Args:
            transport (picows.WSTransport): WebSocket transport instance
        """
        if self._specific_ping_msg:
            transport.send(WSMsgType.TEXT, self._specific_ping_msg)
            self._log.debug(
                f"Sent user specific ping: `{self._specific_ping_msg.decode()}`."
            )
        else:
            transport.send_ping()
            self._log.debug("Sent default ping.")

    def on_ws_connected(self, transport: WSTransport) -> None:
        """Called when WebSocket connection is established.

        Args:
            transport (picows.WSTransport): WebSocket transport instance
        """
        self._log.debug("Connected to Websocket...")

    def on_ws_disconnected(self, transport: WSTransport) -> None:
        """Called when WebSocket connection is closed.

        Args:
            transport (picows.WSTransport): WebSocket transport instance
        """
        self._log.debug("Disconnected from Websocket.")

    def on_ws_frame(self, transport: WSTransport, frame: WSFrame) -> None:
        """Handle incoming WebSocket frames.

        Args:
            transport (picows.WSTransport): WebSocket transport instance
            frame (picows.WSFrame): Received WebSocket frame
        """
        try:
            match frame.msg_type:
                case WSMsgType.PING:
                    # Only send pong if auto_pong is disabled
                    self._log.debug("Received PING frame, sending PONG frame...")
                    transport.send_pong(frame.get_payload_as_bytes())
                    return
                case WSMsgType.TEXT:
                    # Queue raw bytes for handler to decode
                    self._callback(frame.get_payload_as_bytes())
                    return
                case WSMsgType.CLOSE:
                    close_code = frame.get_close_code()
                    close_msg = frame.get_close_message()
                    self._log.warning(
                        f"Received close frame. Close code: {close_code}, Close message: {close_msg.decode()}"
                    )
                    return
        except Exception as e:
            import traceback

            self._log.error(
                f"Error processing message: {str(e)}\nTraceback: {traceback.format_exc()}\nws_frame: {frame.get_payload_as_utf8_text()}"
            )


class WSClient(ABC):
    def __init__(
        self,
        url: str,
        handler: Callable[..., Any],
        task_manager: TaskManager,
        clock: LiveClock,
        specific_ping_msg: bytes = None,
        reconnect_interval: int = 1,
        ping_idle_timeout: int = 2,
        ping_reply_timeout: int = 1,
        auto_ping_strategy: Literal[
            "ping_when_idle", "ping_periodically"
        ] = "ping_when_idle",
        enable_auto_ping: bool = True,
        enable_auto_pong: bool = False,
    ):
        self._clock = clock
        self._url = url
        self._specific_ping_msg = specific_ping_msg
        self._reconnect_interval = reconnect_interval
        self._ping_idle_timeout = ping_idle_timeout
        self._ping_reply_timeout = ping_reply_timeout
        self._enable_auto_pong = enable_auto_pong
        self._enable_auto_ping = enable_auto_ping
        self._listener: Listener = None
        self._transport = None
        self._subscriptions = []
        self._callback = handler
        if auto_ping_strategy == "ping_when_idle":
            self._auto_ping_strategy = WSAutoPingStrategy.PING_WHEN_IDLE
        elif auto_ping_strategy == "ping_periodically":
            self._auto_ping_strategy = WSAutoPingStrategy.PING_PERIODICALLY
        self._task_manager = task_manager
        self._log = Logger(name=type(self).__name__)

    @property
    def connected(self):
        return self._transport and self._listener

    async def _connect(self):
        self._log.debug(f"Connecting to Websocket at {self._url}...")
        WSListenerFactory = lambda: Listener(  # noqa: E731
            self._callback, self._log, self._specific_ping_msg
        )
        self._transport, self._listener = await ws_connect(
            WSListenerFactory,
            self._url,
            enable_auto_ping=self._enable_auto_ping,
            auto_ping_idle_timeout=self._ping_idle_timeout,
            auto_ping_reply_timeout=self._ping_reply_timeout,
            auto_ping_strategy=self._auto_ping_strategy,
            enable_auto_pong=self._enable_auto_pong,
        )

    async def connect(self):
        if not self.connected:
            await self._connect()
            self._task_manager.create_task(self._connection_handler())

    async def _connection_handler(self):
        while True:
            try:
                if not self.connected:
                    await self._connect()
                    await self._resubscribe()
                await self._transport.wait_disconnected()
            except Exception as e:
                self._log.error(f"Connection error: {e}")

            if self.connected:
                self._log.warning("Websocket reconnecting...")
                self.disconnect()
            await asyncio.sleep(self._reconnect_interval)

    def _send(self, payload: dict):
        self._transport.send(WSMsgType.TEXT, msgspec.json.encode(payload))

    def disconnect(self):
        if self.connected:
            self._log.debug("Disconnecting from websocket...")
            self._transport.disconnect()
            self._transport, self._listener = None, None

    @abstractmethod
    async def _resubscribe(self):
        pass



"""Public interface for strategy web integration."""

from __future__ import annotations

from .app import StrategyFastAPI, create_strategy_app
from .server import StrategyWebServer

# Default app instance for convenience imports (``from nexustrader.web import app``)
app = create_strategy_app()

__all__ = [
    "StrategyFastAPI",
    "StrategyWebServer",
    "create_strategy_app",
    "app",
]



import msgspec
from decimal import Decimal
from typing import Any, Dict, List
from nexustrader.schema import BaseMarket, Balance, BookOrderData
from nexustrader.exchange.binance.constants import (
    BinanceAccountEventReasonType,
    BinanceOrderStatus,
    BinanceOrderType,
    BinancePositionSide,
    BinanceWsEventType,
    BinanceKlineInterval,
    BinancePriceMatch,
    BinanceUserDataStreamWsEventType,
    BinanceOrderSide,
    BinanceTimeInForce,
    BinanceExecutionType,
    BinanceFuturesWorkingType,
    BinanceBusinessUnit,
)


class BinanceResultId(msgspec.Struct):
    id: int | None = None


class BinanceFuturesBalanceInfo(msgspec.Struct, frozen=True):
    asset: str  # asset name
    walletBalance: str  # wallet balance
    unrealizedProfit: str  # unrealized profit
    marginBalance: str  # margin balance
    maintMargin: str  # maintenance margin required
    initialMargin: str  # total initial margin required with current mark price
    positionInitialMargin: (
        str  # initial margin required for positions with current mark price
    )
    openOrderInitialMargin: (
        str  # initial margin required for open orders with current mark price
    )
    crossWalletBalance: str  # crossed wallet balance
    crossUnPnl: str  # unrealized profit of crossed positions
    availableBalance: str  # available balance
    maxWithdrawAmount: str  # maximum amount for transfer out
    # whether the asset can be used as margin in Multi - Assets mode
    marginAvailable: bool | None = None
    updateTime: int | None = None  # last update time

    def parse_to_balance(self) -> Balance:
        free = Decimal(self.availableBalance)
        locked = Decimal(self.marginBalance) - free
        return Balance(
            asset=self.asset,
            free=free,
            locked=locked,
        )


class BinanceFuturesPositionInfo(msgspec.Struct, kw_only=True):
    symbol: str  # symbol name
    initialMargin: str  # initial margin required with current mark price
    maintMargin: str  # maintenance margin required
    unrealizedProfit: str  # unrealized profit
    positionInitialMargin: (
        str  # initial margin required for positions with current mark price
    )
    openOrderInitialMargin: (
        str  # initial margin required for open orders with current mark price
    )
    leverage: str  # current initial leverage
    isolated: bool  # if the position is isolated
    entryPrice: str  # average entry price
    maxNotional: str | None = None  # maximum available notional with current leverage
    bidNotional: str | None = None  # bids notional, ignore
    askNotional: str | None = None  # ask notional, ignore
    positionSide: BinancePositionSide  # position side
    positionAmt: str  # position amount
    updateTime: int
    breakEvenPrice: str | None = None  # break-even price
    maxQty: str | None = None  # maximum quantity of base asset


class BinancePortfolioMarginPositionRisk(msgspec.Struct, kw_only=True):
    """
    "entryPrice": "0.00000",
    "leverage": "10",
    "markPrice": "6679.50671178",
    "maxNotionalValue": "20000000",
    "positionAmt": "0.000",
    "notional": "0",
    "symbol": "BTCUSDT",
    "unRealizedProfit": "0.00000000",
    "liquidationPrice": "6170.20509059",
    "positionSide": "BOTH",
    "updateTime": 1625474304765

    "symbol": "BTCUSD_201225",
    "positionAmt": "1",
    "entryPrice": "11707.70000003",
    "markPrice": "11788.66626667",
    "unRealizedProfit": "0.00005866",
    "liquidationPrice": "6170.20509059",
    "leverage": "125",
    "positionSide": "LONG",
    "updateTime": 1627026881327,
    "maxQty": "50",
    "notionalValue": "0.00084827"
    """

    symbol: str
    markPrice: str
    entryPrice: str
    unRealizedProfit: str
    positionAmt: str
    positionSide: BinancePositionSide
    liquidationPrice: str
    updateTime: int
    leverage: str

    notional: str | None = None
    maxNotionalValue: str | None = None

    maxQty: str | None = None
    notionalValue: str | None = None


class BinancePortfolioMarginBalance(msgspec.Struct, kw_only=True):
    """
    "asset": "USDT",    // asset name
    "totalWalletBalance": "122607.35137903", // wallet balance =  cross margin free + cross margin locked + UM wallet balance + CM wallet balance
    "crossMarginAsset": "92.27530794", // crossMarginAsset = crossMarginFree + crossMarginLocked
    "crossMarginBorrowed": "10.00000000", // principal of cross margin
    "crossMarginFree": "100.00000000", // free asset of cross margin
    "crossMarginInterest": "0.72469206", // interest of cross margin
    "crossMarginLocked": "3.00000000", //lock asset of cross margin
    "umWalletBalance": "0.00000000",  // wallet balance of um
    "umUnrealizedPNL": "23.72469206",     // unrealized profit of um
    "cmWalletBalance": "23.72469206",       // wallet balance of cm
    "cmUnrealizedPNL": "",    // unrealized profit of cm
    "updateTime": 1617939110373,
    "negativeBalance": "0"
    """

    asset: str
    totalWalletBalance: str
    crossMarginAsset: str
    crossMarginBorrowed: str
    crossMarginFree: str
    crossMarginInterest: str
    crossMarginLocked: str
    umWalletBalance: str
    umUnrealizedPNL: str
    cmWalletBalance: str
    cmUnrealizedPNL: str
    updateTime: int
    negativeBalance: str

    def parse_to_balance(self) -> Balance:
        return Balance(
            asset=self.asset,
            free=Decimal(self.crossMarginFree)
            + Decimal(self.umWalletBalance)
            + Decimal(self.cmWalletBalance),
            locked=Decimal(self.crossMarginLocked),
        )


class BinanceFuturesAccountInfo(msgspec.Struct, kw_only=True):
    feeTier: int  # account commission tier
    canTrade: bool  # if can trade
    canDeposit: bool  # if can transfer in asset
    canWithdraw: bool  # if can transfer out asset
    updateTime: int
    totalInitialMargin: str | None = (
        None  # total initial margin required with current mark price (useless with isolated positions), only for USDT
    )
    totalMaintMargin: str | None = (
        None  # total maintenance margin required, only for USDT asset
    )
    totalWalletBalance: str | None = None  # total wallet balance, only for USDT asset
    totalUnrealizedProfit: str | None = (
        None  # total unrealized profit, only for USDT asset
    )
    totalMarginBalance: str | None = None  # total margin balance, only for USDT asset
    # initial margin required for positions with current mark price, only for USDT asset
    totalPositionInitialMargin: str | None = None
    # initial margin required for open orders with current mark price, only for USDT asset
    totalOpenOrderInitialMargin: str | None = None
    totalCrossWalletBalance: str | None = (
        None  # crossed wallet balance, only for USDT asset
    )
    # unrealized profit of crossed positions, only for USDT asset
    totalCrossUnPnl: str | None = None
    availableBalance: str | None = None  # available balance, only for USDT asset
    maxWithdrawAmount: str | None = (
        None  # maximum amount for transfer out, only for USDT asset
    )
    assets: list[BinanceFuturesBalanceInfo]
    positions: list[BinanceFuturesPositionInfo]

    def parse_to_balances(self) -> List[Balance]:
        return [balance.parse_to_balance() for balance in self.assets]


class BinanceSpotBalanceInfo(msgspec.Struct):
    asset: str
    free: str
    locked: str

    def parse_to_balance(self) -> Balance:
        return Balance(
            asset=self.asset,
            free=Decimal(self.free),
            locked=Decimal(self.locked),
        )


class BinanceSpotAccountInfo(msgspec.Struct, frozen=True):
    makerCommission: int
    takerCommission: int
    buyerCommission: int
    sellerCommission: int
    canTrade: bool
    canWithdraw: bool
    canDeposit: bool
    updateTime: int
    accountType: str
    balances: list[BinanceSpotBalanceInfo]
    permissions: list[str]

    def parse_to_balances(self) -> List[Balance]:
        return [balance.parse_to_balance() for balance in self.balances]


class BinanceSpotOrderUpdateMsg(msgspec.Struct, kw_only=True):
    C: str | None = None
    E: int
    F: str | None = None
    I: int  # noqa: E741
    L: str
    M: bool | None = None
    N: str | None = None
    O: int  # noqa: E741
    P: str
    Q: str | None = None
    S: BinanceOrderSide
    T: int
    V: str
    W: int | None = None  # Working Time
    X: BinanceOrderStatus
    Y: str
    Z: str
    c: str
    e: BinanceUserDataStreamWsEventType
    f: BinanceTimeInForce
    g: int
    i: int
    l: str  # noqa: E741
    m: bool
    n: str | None = None
    o: BinanceOrderType
    p: str
    q: str
    r: str | None = None
    s: str
    t: int
    v: int | None = None
    w: bool
    x: BinanceExecutionType
    z: str

    A: str | None = None  # Prevented Quantity
    B: str | None = None  # Last Prevented Quantity
    D: int | None = None  # trailing time
    J: int | None = None  # strategy type
    U: int | None = None  # CounterOrderId
    d: int | None = None  # trailing Delta
    j: int | None = None  # strategy id
    u: int | None = None  # Trade Group id


class BinanceFuturesOrderData(msgspec.Struct, kw_only=True):
    L: str  # Last Filled Price
    N: str  # Commission Asset
    R: bool  # Is reduce only
    S: BinanceOrderSide
    T: int  # Order Trade Time
    V: str  # Order Filled Accumulated Quantity
    X: BinanceOrderStatus
    a: str  # Ask Notional
    ap: str  # Average Price
    b: str  # Bids Notional
    c: str  # Client Order ID
    f: BinanceTimeInForce
    i: int  # Order ID
    l: str  # Order Last Filled Quantity # noqa: E741
    m: bool  # Is trade the maker side
    n: str  # Commission, will not push if no commission
    o: BinanceOrderType
    p: str  # Original Price
    ps: BinancePositionSide
    q: str  # Original Quantity
    rp: str  # Realized Profit of the trade
    s: str  # Symbol
    sp: str  # Stop Price
    t: int  # Trade ID
    x: BinanceExecutionType
    z: str  # Order Filled Accumulated Quantity

    AP: str | None = None  # Activation Price
    cp: bool | None = None  # If Close-All, pushed with conditional order
    cr: str | None = None  # Callback Rate
    gtd: int | None = None  # TIF GTD order auto cancel time
    ot: BinanceOrderType | None = None
    pP: bool | None = None  # ignore
    pm: str | None = None  # Order Margin Type
    si: int | None = None  # ignore
    ss: int | None = None  # ignore
    wt: BinanceFuturesWorkingType | None = None
    ma: str | None = None  # Order Margin Type
    st: str | None = (
        None  #  Strategy type, only pushed with conditional order triggered
    )


class BinanceFuturesOrderUpdateMsg(msgspec.Struct, kw_only=True):
    """
    WebSocket message for Binance Futures Order Update events.
    """

    e: BinanceUserDataStreamWsEventType
    E: int  # Event Time
    T: int  # Transaction Time
    fs: BinanceBusinessUnit | None = (
        None  # Event business unit. 'UM' for USDS-M futures and 'CM' for COIN-M futures
    )
    o: BinanceFuturesOrderData


class BinanceMarkPriceDataStream(msgspec.Struct):
    e: BinanceWsEventType
    E: int
    s: str
    p: str
    i: str
    P: str
    r: str
    T: int


class BinanceMarkPrice(msgspec.Struct):
    data: BinanceMarkPriceDataStream
    stream: str


class BinanceKlineData(msgspec.Struct):
    t: int  # Kline start time
    T: int  # Kline close time
    s: str  # Symbol
    i: BinanceKlineInterval  # Interval
    f: int  # First trade ID
    L: int  # Last trade ID
    o: str  # Open price
    c: str  # Close price
    h: str  # High price
    l: str  # Low price # noqa
    v: str  # Base asset volume
    n: int  # Number of trades
    x: bool  # Is this kline closed?
    q: str  # Quote asset volume
    V: str  # Taker buy base asset volume
    Q: str  # Taker buy quote asset volume
    B: str  # Ignore


class BinanceKlineDataStream(msgspec.Struct):
    e: BinanceWsEventType
    E: int
    s: str
    k: BinanceKlineData


class BinanceKline(msgspec.Struct):
    data: BinanceKlineDataStream
    stream: str


class BinanceTradeDataStream(msgspec.Struct):
    e: BinanceWsEventType
    E: int
    s: str
    t: int
    p: str
    q: str
    T: int
    m: bool # Is the buyer the market maker? true -> side=SELL, false -> side=BUY


class BinanceTradeData(msgspec.Struct):
    data: BinanceTradeDataStream
    stream: str


class BinanceSpotBookTickerData(msgspec.Struct):
    """
      {
        "u":400900217,     // order book updateId
        "s":"BNBUSDT",     // symbol
        "b":"25.35190000", // best bid price
        "B":"31.21000000", // best bid qty
        "a":"25.36520000", // best ask price
        "A":"40.66000000"  // best ask qty
    }
    """

    u: int
    s: str
    b: str
    B: str
    a: str
    A: str


class BinanceSpotBookTicker(msgspec.Struct):
    data: BinanceSpotBookTickerData
    stream: str


class BinanceFuturesBookTickerData(msgspec.Struct):
    e: BinanceWsEventType
    u: int
    E: int
    T: int
    s: str
    b: str
    B: str
    a: str
    A: str


class BinanceFuturesBookTicker(msgspec.Struct):
    data: BinanceFuturesBookTickerData
    stream: str


class BinanceWsMessageGeneralData(msgspec.Struct):
    e: BinanceWsEventType | None = None
    u: int | None = None


class BinanceWsMessageGeneral(msgspec.Struct):
    data: BinanceWsMessageGeneralData


class BinanceUserDataStreamMsg(msgspec.Struct):
    e: BinanceUserDataStreamWsEventType | None = None


class BinanceListenKey(msgspec.Struct):
    listenKey: str


class BinanceUserTrade(msgspec.Struct, frozen=True):
    commission: str
    commissionAsset: str
    price: str
    qty: str

    # Parameters not present in 'fills' list (see FULL response of BinanceOrder)
    symbol: str | None = None
    id: int | None = None
    orderId: int | None = None
    time: int | None = None
    quoteQty: str | None = None  # SPOT/MARGIN & USD-M FUTURES only

    # Parameters in SPOT/MARGIN only:
    orderListId: int | None = None  # unless OCO, the value will always be -1
    isBuyer: bool | None = None
    isMaker: bool | None = None
    isBestMatch: bool | None = None
    tradeId: int | None = None  # only in BinanceOrder FULL response

    # Parameters in FUTURES only:
    buyer: bool | None = None
    maker: bool | None = None
    realizedPnl: str | None = None
    side: BinanceOrderSide | None = None
    positionSide: str | None = None
    baseQty: str | None = None  # COIN-M FUTURES only
    pair: str | None = None  # COIN-M FUTURES only


class BinanceOrder(msgspec.Struct, frozen=True):
    symbol: str
    orderId: int
    clientOrderId: str

    # Parameters not in ACK response:
    price: str | None = None
    origQty: str | None = None
    executedQty: str | None = None
    status: BinanceOrderStatus | None = None
    timeInForce: BinanceTimeInForce | None = None
    goodTillDate: int | None = None
    type: BinanceOrderType | None = None
    side: BinanceOrderSide | None = None
    stopPrice: str | None = (
        None  # please ignore when order type is TRAILING_STOP_MARKET
    )
    time: int | None = None
    updateTime: int | None = None

    # Parameters in SPOT/MARGIN only:
    orderListId: int | None = None  # Unless OCO, the value will always be -1
    cumulativeQuoteQty: str | None = None  # cumulative quote qty
    icebergQty: str | None = None
    isWorking: bool | None = None
    workingTime: int | None = None
    origQuoteOrderQty: str | None = None
    selfTradePreventionMode: str | None = None
    transactTime: int | None = None  # POST & DELETE methods only
    fills: list[BinanceUserTrade] | None = None  # FULL response only

    # Parameters in FUTURES only:
    avgPrice: str | None = None
    origType: BinanceOrderType | None = None
    reduceOnly: bool | None = None
    positionSide: BinancePositionSide | None = None
    closePosition: bool | None = None
    activatePrice: str | None = (
        None  # activation price, only for TRAILING_STOP_MARKET order
    )
    priceRate: str | None = None  # callback rate, only for TRAILING_STOP_MARKET order
    workingType: str | None = None
    priceProtect: bool | None = None  # if conditional order trigger is protected
    cumQuote: str | None = None  # USD-M FUTURES only
    cumBase: str | None = None  # COIN-M FUTURES only
    pair: str | None = None  # COIN-M FUTURES only


class BinanceFuturesModifyOrderResponse(msgspec.Struct, frozen=True, kw_only=True):
    """
        {
            "orderId": 20072994037,
            "symbol": "BTCUSDT",
            "pair": "BTCUSDT",
            "status": "NEW",
            "clientOrderId": "LJ9R4QZDihCaS8UAOOLpgW",
            "price": "30005",
            "avgPrice": "0.0",
            "origQty": "1",
            "executedQty": "0",
            "cumQty": "0",
            "cumBase": "0",
            "timeInForce": "GTC",
            "type": "LIMIT",
            "reduceOnly": false,
            "closePosition": false,
            "side": "BUY",
            "positionSide": "LONG",
            "stopPrice": "0",
            "workingType": "CONTRACT_PRICE",
            "priceProtect": false,
            "origType": "LIMIT",
        "priceMatch": "NONE",              //price match mode
        "selfTradePreventionMode": "NONE", //self trading preventation mode
        "goodTillDate": 0,                 //order pre-set auot cancel time for TIF GTD order
        "updateTime": 1629182711600
    }
    """

    orderId: int
    symbol: str
    pair: str | None = None
    status: BinanceOrderStatus
    clientOrderId: str
    price: str
    avgPrice: str
    origQty: str
    executedQty: str
    cumQty: str
    cumBase: str | None = None
    cumQuote: str | None = None
    timeInForce: BinanceTimeInForce
    type: BinanceOrderType
    reduceOnly: bool | None = None
    closePosition: bool | None = None
    side: BinanceOrderSide
    positionSide: BinancePositionSide
    stopPrice: str | None = None
    workingType: BinanceFuturesWorkingType | None = None
    priceProtect: bool | None = None
    origType: BinanceOrderType
    priceMatch: BinancePriceMatch | None = None
    selfTradePreventionMode: str | None = None
    goodTillDate: int | None = None
    updateTime: int


class BinanceMarketInfo(msgspec.Struct):
    symbol: str = None
    status: str = None
    baseAsset: str = None
    baseAssetPrecision: str | int = None
    quoteAsset: str = None
    quotePrecision: str | int = None
    quoteAssetPrecision: str | int = None
    baseCommissionPrecision: str | int = None
    quoteCommissionPrecision: str | int = None
    orderTypes: List[BinanceOrderType] = None
    icebergAllowed: bool = None
    ocoAllowed: bool = None
    otoAllowed: bool = None
    quoteOrderQtyMarketAllowed: bool = None
    allowTrailingStop: bool = None
    cancelReplaceAllowed: bool = None
    isSpotTradingAllowed: bool = None
    isMarginTradingAllowed: bool = None
    filters: List[Dict[str, Any]] = None
    permissions: List[str] = None
    permissionSets: List[List[str] | str] = None
    defaultSelfTradePreventionMode: str = None
    allowedSelfTradePreventionModes: List[str] = None


class BinanceMarket(BaseMarket):
    info: BinanceMarketInfo
    feeSide: str


class BinanceFuturesBalanceData(msgspec.Struct):
    a: str
    wb: str  # wallet balance
    cw: str  # cross wallet balance
    bc: str  # wallet change except PnL and Commission

    def parse_to_balance(self) -> Balance:
        return Balance(
            asset=self.a,
            free=Decimal(self.wb),
            locked=Decimal(0),
        )


class BinanceFuturesPositionData(msgspec.Struct, kw_only=True):
    s: str
    pa: str  # position amount
    ep: str  # entry price
    bep: str | float  # breakeven price
    cr: str  # (Pre-fee) Accumulated Realized
    up: str  # Unrealized PnL
    mt: str | None = None  # margin type (if isolated position)
    iw: str | None = None  # isolated wallet (if isolated position)
    ps: BinancePositionSide


class BinanceFuturesUpdateData(msgspec.Struct, kw_only=True):
    m: BinanceAccountEventReasonType
    B: list[BinanceFuturesBalanceData]
    P: list[BinanceFuturesPositionData]

    def parse_to_balances(self) -> List[Balance]:
        return [balance.parse_to_balance() for balance in self.B]


class BinanceFuturesUpdateMsg(msgspec.Struct, kw_only=True):
    e: BinanceUserDataStreamWsEventType
    E: int
    T: int
    fs: BinanceBusinessUnit | None = None
    a: BinanceFuturesUpdateData


class BinanceSpotBalanceData(msgspec.Struct):
    a: str  # asset
    f: str  # free
    l: str  # locked # noqa: E741

    def parse_to_balance(self) -> Balance:
        return Balance(
            asset=self.a,
            free=Decimal(self.f),
            locked=Decimal(self.l),
        )


class BinanceSpotUpdateMsg(msgspec.Struct, kw_only=True):
    e: BinanceUserDataStreamWsEventType  # event type
    E: int  # event time
    u: int  # Time of last account update
    B: list[BinanceSpotBalanceData]  # balance array of the account

    def parse_to_balances(self) -> List[Balance]:
        return [balance.parse_to_balance() for balance in self.B]


class BinanceResponseKline(msgspec.Struct, array_like=True):
    """
    [
        1499040000000,      // Kline open time
        "0.01634790",       // Open price
        "0.80000000",       // High price
        "0.01575800",       // Low price
        "0.01577100",       // Close price
        "148976.11427815",  // Volume
        1499644799999,      // Kline Close time
        "2434.19055334",    // Quote asset volume
        308,                // Number of trades
        "1756.87402397",    // Taker buy base asset volume
        "28.46694368",      // Taker buy quote asset volume
        "0"                 // Unused field, ignore.
    ]
    """

    open_time: int
    open: str
    high: str
    low: str
    close: str
    volume: str
    close_time: int
    asset_volume: str
    trades_count: int
    taker_base_volume: str
    taker_quote_volume: str
    ignore: str


class BinanceIndexResponseKline(msgspec.Struct, array_like=True):
    """
    1591256400000,      	// Open time
    "9653.69440000",    	// Open
    "9653.69640000",     	// High
    "9651.38600000",     	// Low
    "9651.55200000",     	// Close (or latest price)
    "0	", 					// Ignore
    1591256459999,      	// Close time
    "0",    				// Ignore
    60,                		// Ignore
    "0",    				// Ignore
    "0",      				// Ignore
    "0" 					// Ignore
    """

    open_time: int
    open: str
    high: str
    low: str
    close: str
    ignore_1: str
    close_time: int
    ignore_2: str
    ignore_3: int
    ignore_4: str
    ignore_5: str
    ignore_6: str


class BinanceSpotOrderBookMsg(msgspec.Struct):
    """
    WebSocket message for 'Binance Spot/Margin' Partial Book Depth Streams.
    """

    stream: str
    data: "BinanceSpotOrderBookData"


class BinanceSpotOrderBookData(msgspec.Struct):
    """
    Websocket message 'inner struct' for 'Binance Spot/Margin Partial Book Depth
    Streams.'.
    """

    lastUpdateId: int
    bids: list["BinanceOrderBookDelta"]
    asks: list["BinanceOrderBookDelta"]


class BinanceOrderBookDelta(msgspec.Struct, array_like=True):
    """
    Schema of single ask/bid delta.
    """

    price: str
    size: str

    def parse_to_book_order_data(self) -> BookOrderData:
        return BookOrderData(
            price=float(self.price),
            size=float(self.size),
        )


class BinanceFuturesOrderBookMsg(msgspec.Struct, frozen=True):
    """
    WebSocket message from Binance Partial & Diff.

    Book Depth Streams.

    """

    stream: str
    data: "BinanceFuturesOrderBookData"


class BinanceFuturesOrderBookData(msgspec.Struct, frozen=True):
    """
    WebSocket message 'inner struct' for Binance Partial & Diff.

    Book Depth Streams.

    """

    e: str  # Event type
    E: int  # Event time
    s: str  # Symbol
    U: int  # First update ID in event
    u: int  # Final update ID in event
    b: list[BinanceOrderBookDelta]  # Bids to be updated
    a: list[BinanceOrderBookDelta]  # Asks to be updated

    T: int | None = None  # FUTURES only, transaction time
    pu: int | None = None  # FUTURES only, previous final update ID
    ps: str | None = None  # COIN-M FUTURES only, pair


class BinanceCancelAllOrdersResponse(msgspec.Struct, frozen=True):
    code: int
    msg: str


class BinanceFundingRateResponse(msgspec.Struct, frozen=True):
    symbol: str
    fundingRate: str
    fundingTime: int
    markPrice: str | None = None


class BinanceBatchOrderResponse(msgspec.Struct, frozen=True, omit_defaults=True):
    clientOrderId: str | None = None
    cumQty: str | None = None
    executedQty: str | None = None
    orderId: int | None = None
    avgPrice: str | None = None
    origQty: str | None = None
    price: str | None = None
    reduceOnly: bool | None = None
    side: BinanceOrderSide | None = None
    positionSide: BinancePositionSide | None = None
    status: BinanceOrderStatus | None = None
    stopPrice: str | None = None
    closePosition: bool | None = None
    symbol: str | None = None
    timeInForce: BinanceTimeInForce | None = None
    type: BinanceOrderType | None = None
    origType: BinanceOrderType | None = None
    updateTime: int | None = None
    priceProtect: bool | None = None
    # USD-M specific fields
    cumQuote: str | None = None
    goodTillDate: int | None = None
    # Coin-M specific fields
    cumBase: str | None = None
    pair: str | None = None
    # Optional fields for both
    activatePrice: str | None = None
    priceRate: str | None = None
    workingType: BinanceFuturesWorkingType | None = None
    priceMatch: BinancePriceMatch | None = None
    selfTradePreventionMode: str | None = None
    code: int | None = None
    msg: str | None = None


################################################################################
# GET /fapi/v1/ticker/24hr and GET /dapi/v1/ticker/24hr
################################################################################


class BinanceFuture24hrTicker(msgspec.Struct):
    """
    24hr ticker data structure for Binance Futures (FAPI and DAPI).
    Supports both USD-M and Coin-M futures.
    """

    symbol: str  # Symbol name
    priceChange: str  # Price change
    priceChangePercent: str  # Price change percentage
    weightedAvgPrice: str  # Weighted average price
    lastPrice: str  # Last price
    lastQty: str  # Last quantity
    openPrice: str  # Open price
    highPrice: str  # High price
    lowPrice: str  # Low price
    volume: str  # Volume
    openTime: int  # Open time
    closeTime: int  # Close time
    firstId: int  # First trade ID
    lastId: int  # Last trade ID
    count: int  # Trade count

    # FAPI specific fields
    quoteVolume: str | None = None  # Quote volume (FAPI only)

    # DAPI specific fields
    pair: str | None = None  # Pair name (DAPI only)
    baseVolume: str | None = None  # Base volume (DAPI only)


class BinanceSpot24hrTicker(msgspec.Struct):
    """
    24hr ticker data structure for Binance Spot.
    Supports both FULL and MINI response types.
    """

    symbol: str  # Symbol name
    openPrice: str  # Opening price
    highPrice: str  # Highest price
    lowPrice: str  # Lowest price
    lastPrice: str  # Last price
    volume: str  # Total trade volume (base asset)
    quoteVolume: str  # Total trade volume (quote asset)
    openTime: int  # Start of ticker interval
    closeTime: int  # End of ticker interval
    firstId: int  # First trade ID
    lastId: int  # Last trade ID
    count: int  # Total trade count

    # FULL response additional fields
    priceChange: str | None = None  # Price change
    priceChangePercent: str | None = None  # Price change percentage
    weightedAvgPrice: str | None = None  # Weighted average price
    prevClosePrice: str | None = None  # Previous close price
    lastQty: str | None = None  # Last quantity
    bidPrice: str | None = None  # Best bid price
    bidQty: str | None = None  # Best bid quantity
    askPrice: str | None = None  # Best ask price
    askQty: str | None = None  # Best ask quantity


class BinanceWsOrderResponseResult(msgspec.Struct, frozen=True):
    orderId: int
    symbol: str
    clientOrderId: str


class BinanceWsOrderResponseError(msgspec.Struct, frozen=True):
    code: int
    msg: str

    @property
    def format_str(self) -> str:
        return f"code={self.code} error={self.msg}"


class BinanceWsOrderResponse(msgspec.Struct, frozen=True):
    id: str
    status: int
    result: BinanceWsOrderResponseResult | None = None
    error: BinanceWsOrderResponseError | None = None

    @property
    def is_success(self) -> bool:
        return self.status == 200

    @property
    def is_failed(self) -> bool:
        return self.status != 200



from enum import Enum
from nexustrader.constants import AccountType
from throttled.asyncio import Throttled, rate_limiter
from throttled import Throttled as ThrottledSync
from throttled import rate_limiter as rate_limiter_sync

from nexustrader.constants import (
    OrderStatus,
    PositionSide,
    OrderSide,
    TimeInForce,
    OrderType,
    KlineInterval,
    RateLimiter,
    RateLimiterSync,
)
from nexustrader.error import KlineSupportedError


class BitgetAccountType(AccountType):
    UTA = "UTA"
    SPOT = "SPOT"
    FUTURE = "FUTURE"
    UTA_DEMO = "UTA_DEMO"
    SPOT_DEMO = "SPOT_DEMO"
    FUTURE_DEMO = "FUTURE_DEMO"
    SPOT_MOCK = "SPOT_MOCK"
    LINEAR_MOCK = "LINEAR_MOCK"
    INVERSE_MOCK = "INVERSE_MOCK"

    @property
    def exchange_id(self):
        return "bitget"

    @property
    def is_testnet(self):
        return self in (
            self.UTA_DEMO,
            self.SPOT_DEMO,
            self.FUTURE_DEMO,
        )

    @property
    def stream_url(self):
        version = "v3" if self.is_uta else "v2"
        if self.is_testnet:
            return f"wss://wspap.bitget.com/{version}/ws"
        else:
            return f"wss://ws.bitget.com/{version}/ws"

    @property
    def is_spot(self):
        return self in (self.SPOT, self.SPOT_DEMO, self.SPOT_MOCK)

    @property
    def is_future(self):
        return self in (
            self.FUTURE,
            self.FUTURE_DEMO,
            self.LINEAR_MOCK,
            self.INVERSE_MOCK,
        )

    @property
    def is_uta(self):
        return self in (self.UTA, self.UTA_DEMO)

    @property
    def is_mock(self):
        return self in (self.SPOT_MOCK, self.LINEAR_MOCK, self.INVERSE_MOCK)

    @property
    def is_linear_mock(self):
        return self == self.LINEAR_MOCK

    @property
    def is_inverse_mock(self):
        return self == self.INVERSE_MOCK

    @property
    def is_spot_mock(self):
        return self == self.SPOT_MOCK


class BitgetInstType(Enum):
    SPOT = "SPOT"
    USDT_FUTURES = "USDT-FUTURES"
    COIN_FUTURES = "COIN-FUTURES"
    USDC_FUTURES = "USDC-FUTURES"
    # SUSDT_FUTURES = "SUSD-FUTURES" # deprecated
    # SUSDC_FUTURES = "SUSDC-FUTURES" # deprecated
    # SCOIN_FUTURES = "SCOIN-FUTURES" # deprecated

    @property
    def is_spot(self) -> bool:
        return self == BitgetInstType.SPOT

    @property
    def is_swap(self) -> bool:
        return not self.is_spot

    @property
    def is_linear(self) -> bool:
        return self in (BitgetInstType.USDT_FUTURES, BitgetInstType.USDC_FUTURES)

    @property
    def is_inverse(self) -> bool:
        return self == BitgetInstType.COIN_FUTURES

    @property
    def is_usdc_swap(self) -> bool:
        return self == BitgetInstType.USDC_FUTURES

    @property
    def is_usdt_swap(self) -> bool:
        return self == BitgetInstType.USDT_FUTURES


class BitgetUtaInstType(Enum):
    SPOT = "spot"
    USDT_FUTURES = "usdt-futures"
    COIN_FUTURES = "coin-futures"
    USDC_FUTURES = "usdc-futures"
    UTA = "UTA"

    @property
    def is_spot(self) -> bool:
        return self == BitgetInstType.SPOT

    @property
    def is_swap(self) -> bool:
        return not self.is_spot

    @property
    def is_linear(self) -> bool:
        return self in (BitgetInstType.USDT_FUTURES, BitgetInstType.USDC_FUTURES)

    @property
    def is_inverse(self) -> bool:
        return self == BitgetInstType.COIN_FUTURES

    @property
    def is_usdc_swap(self) -> bool:
        return self == BitgetInstType.USDC_FUTURES

    @property
    def is_usdt_swap(self) -> bool:
        return self == BitgetInstType.USDT_FUTURES


class BitgetKlineInterval(Enum):
    MINUTE_1 = "1m"
    MINUTE_5 = "5m"
    MINUTE_15 = "15m"
    MINUTE_30 = "30m"
    HOUR_1 = "1H"
    HOUR_4 = "4H"
    HOUR_6 = "6H"
    HOUR_12 = "12H"
    DAY_1 = "1D"


class BitgetOrderSide(Enum):
    BUY = "buy"
    SELL = "sell"


class BitgetOrderStatus(Enum):
    LIVE = "live"
    NEW = "new"
    FILLED = "filled"
    CANCELED = "canceled"
    CANCELLED = "cancelled"  # Alias pointing to the same value
    PARTIALLY_FILLED = "partially_filled"


class BitgetTimeInForce(Enum):
    GTC = "gtc"
    IOC = "ioc"
    FOK = "fok"
    POST_ONLY = "post_only"

    @property
    def is_gtc(self) -> bool:
        return self == BitgetTimeInForce.GTC

    @property
    def is_ioc(self) -> bool:
        return self == BitgetTimeInForce.IOC

    @property
    def is_fok(self) -> bool:
        return self == BitgetTimeInForce.FOK

    @property
    def is_post_only(self) -> bool:
        return self == BitgetTimeInForce.POST_ONLY


class BitgetOrderType(Enum):
    LIMIT = "limit"
    MARKET = "market"

    @property
    def is_limit(self) -> bool:
        return self == BitgetOrderType.LIMIT

    @property
    def is_market(self) -> bool:
        return self == BitgetOrderType.MARKET


class BitgetPositionSide(Enum):
    LONG = "long"
    SHORT = "short"
    NET = "net"  # one-way mode position

    @property
    def is_long(self) -> bool:
        return self == BitgetPositionSide.LONG

    @property
    def is_short(self) -> bool:
        return self == BitgetPositionSide.SHORT

    @property
    def is_net(self) -> bool:
        return self == BitgetPositionSide.NET

    def parse_to_position_side(self) -> PositionSide:
        if self == self.LONG:
            return PositionSide.LONG
        elif self == self.SHORT:
            return PositionSide.SHORT
        else:
            return PositionSide.FLAT


class BitgetEnumParser:
    _kline_interval_map = {
        BitgetKlineInterval.MINUTE_1: KlineInterval.MINUTE_1,
        BitgetKlineInterval.MINUTE_5: KlineInterval.MINUTE_5,
        BitgetKlineInterval.MINUTE_15: KlineInterval.MINUTE_15,
        BitgetKlineInterval.MINUTE_30: KlineInterval.MINUTE_30,
        BitgetKlineInterval.HOUR_1: KlineInterval.HOUR_1,
        BitgetKlineInterval.HOUR_4: KlineInterval.HOUR_4,
        BitgetKlineInterval.HOUR_6: KlineInterval.HOUR_6,
        BitgetKlineInterval.HOUR_12: KlineInterval.HOUR_12,
        BitgetKlineInterval.DAY_1: KlineInterval.DAY_1,
    }

    _order_status_map = {
        BitgetOrderStatus.LIVE: OrderStatus.ACCEPTED,
        BitgetOrderStatus.FILLED: OrderStatus.FILLED,
        BitgetOrderStatus.CANCELED: OrderStatus.CANCELED,
        BitgetOrderStatus.CANCELLED: OrderStatus.CANCELED,  # Alias pointing to the same value
        BitgetOrderStatus.PARTIALLY_FILLED: OrderStatus.PARTIALLY_FILLED,
    }

    _order_side_map = {
        BitgetOrderSide.BUY: OrderSide.BUY,
        BitgetOrderSide.SELL: OrderSide.SELL,
    }

    _time_in_force_map = {
        BitgetTimeInForce.GTC: TimeInForce.GTC,
        BitgetTimeInForce.IOC: TimeInForce.IOC,
        BitgetTimeInForce.FOK: TimeInForce.FOK,
    }

    _order_type_map = {
        BitgetOrderType.LIMIT: OrderType.LIMIT,
        BitgetOrderType.MARKET: OrderType.MARKET,
    }

    _kline_interval_to_bitget_map = {v: k for k, v in _kline_interval_map.items()}
    _order_side_to_bitget_map = {v: k for k, v in _order_side_map.items()}
    _time_in_force_to_bitget_map = {v: k for k, v in _time_in_force_map.items()}

    @classmethod
    def parse_kline_interval(cls, interval: BitgetKlineInterval) -> KlineInterval:
        return cls._kline_interval_map[interval]

    @classmethod
    def to_bitget_kline_interval(cls, interval: KlineInterval) -> BitgetKlineInterval:
        interval = cls._kline_interval_to_bitget_map.get(interval)
        if not interval:
            raise KlineSupportedError(
                f"Unsupported KlineInterval: {interval}. Supported intervals: {list(cls._kline_interval_to_bitget_map.keys())}"
            )
        return interval

    @classmethod
    def parse_order_status(cls, status: BitgetOrderStatus) -> OrderStatus:
        # we do not care the UTA account NEW order status, so we ignore it
        # ref: https://www.bitget.com/zh-CN/api-doc/uta/websocket/private/Order-Channel
        return cls._order_status_map.get(status, None)

    @classmethod
    def parse_order_side(cls, side: BitgetOrderSide) -> OrderSide:
        return cls._order_side_map[side]

    @classmethod
    def to_bitget_order_side(cls, side: OrderSide) -> BitgetOrderSide:
        return cls._order_side_to_bitget_map[side]

    @classmethod
    def parse_time_in_force(cls, tif: BitgetTimeInForce) -> TimeInForce:
        return cls._time_in_force_map[tif]

    @classmethod
    def to_bitget_time_in_force(cls, tif: TimeInForce) -> BitgetTimeInForce:
        return cls._time_in_force_to_bitget_map[tif]

    @classmethod
    def parse_order_type(cls, order_type: BitgetOrderType) -> OrderType:
        return cls._order_type_map[order_type]


class BitgetRateLimiter(RateLimiter):
    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: dict[str, Throttled] = {
            "/api/v2/mix/order/place-order": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v2/spot/trade/place-order": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v2/mix/order/cancel-order": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v2/spot/trade/cancel-order": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v3/trade/place-order": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v3/trade/cancel-order": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v3/trade/cancel-symbol-order": Throttled(
                quota=rate_limiter.per_sec(5),
                timeout=1 if enable_rate_limit else -1,
            ),
        }

    def __call__(self, rate_limit_type: str) -> Throttled:
        return self._throttled[rate_limit_type]


class BitgetRateLimiterSync(RateLimiterSync):
    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: dict[str, ThrottledSync] = {
            "/api/v2/mix/order/place-order": ThrottledSync(
                quota=rate_limiter_sync.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v2/spot/trade/place-order": ThrottledSync(
                quota=rate_limiter_sync.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v2/mix/position/all-position": ThrottledSync(
                quota=rate_limiter_sync.per_sec(10),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v3/market/tickers": ThrottledSync(
                quota=rate_limiter_sync.per_sec(20),
                timeout=1 if enable_rate_limit else -1,
            ),
            "/api/v3/position/current-position": ThrottledSync(
                quota=rate_limiter_sync.per_sec(20),
                timeout=1 if enable_rate_limit else -1,
            ),
        }

    def __call__(self, rate_limit_type: str) -> ThrottledSync:
        return self._throttled[rate_limit_type]



from typing import Any


class OkxHttpError(Exception):
    """
    Define the class for all OKX http specific errors.
    """

    def __init__(self, status_code: int, message: str, headers: dict[str, Any]) -> None:
        """
        Define the base class for all OKX http specific errors.
        """
        super().__init__(message)
        self.status = status_code
        self.message = message
        self.headers = headers


class OkxRequestError(Exception):
    """
    The base class for all OKX specific errors.

    References
    ----------
    https://www.okx.com/docs-v5/en/?python#error-code

    """

    def __init__(self, error_code: int, status_code: int | None, message: str | None):
        super().__init__(message)
        self.code = error_code
        self.status_code = status_code
        self.message = message

    def __repr__(self) -> str:
        return f"{type(self).__name__}(code={self.code}, message='{self.message}')"

    __str__ = __repr__


def retry_check(exc: Exception) -> bool:
    import httpx

    if isinstance(exc, httpx.NetworkError):
        return True

    if isinstance(exc, OkxRequestError):
        return exc.code in [50001, 50013, 50026, 51054, 51149, 51412]
